# Copyright (C) 1998 Tuomas J. Lukka 1999, John Stewart CRC Canada.
# DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
# See the GNU Library General Public License (file COPYING in the distribution)
# for conditions of use and redistribution.

#
# Config system idea from PDL
#

system "perl VRMLC.pm";

$VERSION = "0.39";

BEGIN {
 eval "use 5.004";
 warn "VRML::Browser requires Perl v5.0.4 or later -- go to http://www.perl.com"
  if $@;
}

use Cwd qw(getcwd chdir);

# Remove our config argument
@ARGV = map {
	if(/^CONF=(.*)$/) {
		$vrml_conf_file = $1; ();
	} else {
		$_
	}
} @ARGV;

if(!defined $vrml_conf_file) {
	my $defname = "$ENV{HOME}/.vrml.conf";
	if(-e $defname) {
		$vrml_conf_file = $defname;
	}
}

require './vrml.conf';
%VRML_CONFIG_DIST = %VRML_CONFIG;

if(defined $vrml_conf_file) {
 if(!-e $vrml_conf_file) {
	die("Invalid config file '$vrml_conf_file' specified");
 }
 print "Reading configuration from file '$vrml_conf_file'\n";
 require $vrml_conf_file;
}

for(keys %VRML_CONFIG) {
	if(!exists($VRML_CONFIG_DIST{$_})) {
		die("Invalid key '$_' found in user supplied file '$vrml_conf_file'.
This key may no longer be in use, or someone may have garbled the 'vrml.conf' 
file in the distribution. Please correct and try again");
	}
}

%VRML_CONFIG = (%VRML_CONFIG_DIST, %VRML_CONFIG);

print STDERR "Writing Config.pm\n";
open OUT, ">Config.pm" or die("Couldn't write to Config.pm");
print OUT "# AUTOMATICALLY GENERATED BY TOP-LEVEL MAKEFILE.PL -- DO NOT EDIT
\%VRML::Config = (
";
for(keys %VRML_CONFIG) {
	print OUT "\t$_\t=>\t";
	if(defined $VRML_CONFIG{$_}) {
                print OUT '"',quotemeta($VRML_CONFIG{$_}),'"';
	} else {
		print OUT "undef";
	}
	print OUT ",\n";
}
print OUT "VERSION => \"$VERSION\"\n";
print OUT ");\n";
chomp ($myconv = `which convert`);
chomp ($myxsltproc = `which xsltproc`);
chomp ($pwd = getcwd());
chomp ($java = `which java`);
print OUT qq(\$VRML::Browser::CONVERT = "$myconv";\n);
print OUT qq(\$VRML::Browser::XSLTPROC = "$myxsltproc";\n);
print OUT qq(\$VRML::ENV{FREEWRL_FONTS} = "$pwd/fonts";\n);
print OUT qq(\$VRML::Browser::X3DTOVRMLXSL = "$pwd/x3d/X3dToVrml97.xsl";\n);
print OUT "1;\n";
close OUT;

print "Note that 'convert' cannot be found\n" unless $a ;


use ExtUtils::MakeMaker;
WriteMakefile(
      ## OPTIMIZE => "-g",
	NAME => "VRML::VRMLFunc",
	DISTNAME => "FreeWRL",
	dist => {
		COMPRESS => 'gzip -9', SUFFIX => 'gz',
	},
	dynamic_lib => { 'OTHERLDFLAGS' => $VRML_CONFIG{LDFLAGS} },
    VERSION => $VERSION,
    DEFINE => $VRML_CONFIG{FREEWRL_DEFINE},
    INC => $VRML_CONFIG{FREEWRL_INC},
    EXE_FILES => ['freewrl'],
    MAN1PODS => { 'freewrl' => '$(INST_MAN1DIR)/freewrl.$(MAN1EXT)' },
    PL_FILES => {'freewrl.PL' => 'freewrl'},
	PREREQ_PM => { 'Digest::MD5' =>  "2.09",
					'HTML::Parser' => "2.25",
					'MIME::Base64' => "2.11",
					'URI'          => "1.04",
					'LWP'          => "5.47",
				   },
			  LIBS => $VRML_CONFIG{FREEWRL_LIBS},
   ##OBJECT => 'VRMLFunc.o CFuncs/Polyrep.o CFuncs/NormalCalcs.o CFuncs/Tess.o CFuncs/Textures.o CFuncs/LinearAlgebra.o CFuncs/Collision.o CFuncs/MPEG_Utils.o CFuncs/Text.o CFuncs/readpng.o',
   OBJECT => 'VRMLFunc.o',
);


sub MY::post_constants {
	my $post_const = "";

	$post_const =
q{

vpath %.c .:./CFuncs
vpath %.o .:./CFuncs

CFUNCS_SRC:=$(wildcard ./CFuncs/*.c)
OBJECT+=$(CFUNCS_SRC:%c=%o)

};

	return ($post_const);
}

sub MY::dynamic_lib {
"ARMAYBE = :
OTHERLDFLAGS = $VRML_CONFIG{LDFLAGS} 
INST_DYNAMIC_DEP = 

\$\(INST_DYNAMIC\): \$\(OBJECT\) \$\(MYEXTLIB\) \$\(BOOTSTRAP\) \$\(INST_ARCHAUTODIR\)/.exists \$\(EXPORT_LIST\) \$\(PERL_ARCHIVE\) \$\(PERL_ARCHIVE_AFTER\) \$\(INST_DYNAMIC_DEP\)
	\$\(RM_F\) \$\@
	LD_RUN_PATH=\$\(LD_RUN_PATH\) \$\(LD\) \$\(LDFROM\) \$\(OTHERLDFLAGS\) -o \$\@ \$\(MYEXTLIB\) \$\(PERL_ARCHIVE\) \$\(LDLOADLIBS\) \$\(PERL_ARCHIVE_AFTER\) \$\(EXPORT_LIST\)
	\$\(CHMOD\) \$\(PERM_RWX\) \$\@
"
}

sub MY::postamble {
    ## Add extra make instructions that don't quite fit into the
    ## ExtUtils::MakeMaker framework.

    my @postamble = ();
    push(@postamble,
q{
# --- VRML::VRMLFunc:

VRMLFunc.pm VRMLFunc.xs: VRMLC.pm VRMLFields.pm VRMLNodes.pm VRMLRend.pm VRMLExtrusion.pm 
	$(PERL) VRMLC.pm
});

    if ($VRML_CONFIG{NETSCAPE_INST}) {
	push(@postamble,
q{
# --- Mozilla/Netscape Plugin:

install ::
	@cd Plugin && $(MAKE) install
});
    }

    if ($VRML_CONFIG{JAVA_INST} && $VRML_CONFIG{NETSCAPE_CLASSES}) {
	push(@postamble,
q{
# --- Java Archive

install ::
	@cd java && $(MAKE) install
});
    }

    return(join('', @postamble));
}
